import itertools
import numpy as np
from radbm.utils.time import Chronometer

class MonitorTime(object):
    def __init__(self, gen, eta=1):
        self.gen = gen
        self.eta = eta
        self.chrono = Chronometer()
    
    def __iter__(self):
        return self
    
    def __next__(self):
        self.chrono.start()
        try:
            item = next(self.gen)
        finally:
            self.chrono.stop()
        return item
    
    def get_value(self):
        return self.eta*self.chrono.time()
    
class MonitorNext(object):
    def __init__(self, gen, eta=1):
        self.gen = gen
        self.eta = eta
        self.count = 0
    
    def __iter__(self):
        return self
    
    def __next__(self):
        item = next(self.gen)
        self.count += 1 #does not count when next exit
        return item
    
    def get_value(self):
        return self.eta*self.count

def MatchingOracleCost(N, K, k):
    """
    Returns
    ------
    cost : float
        The expected number of call it take to find k documents out of K in
        a database containing N documents (without replacement)
        
    Notes
    -----
    This function does not check if the inputs are valid i.e. N >= K >= k
    """
    return k*(N+1)/(K+1)

class _SSWR(object):
    #utility object
    #used by SSWR and HaltingSSWR in two different but similar ways.
    def __init__(self, relevant, monitor, N, recall=1, on_duplicate_candidates='raise'):
        valid = {'raise', 'ignore'}
        if on_duplicate_candidates not in valid:
            msg = 'on_duplicate_candidate must be in {}, got {}'
            raise ValueError(msg.format(valid, on_duplicate_candidates))
        #const
        self.N = N
        self.recall = recall
        self.relevant = relevant
        self.on_duplicate_candidates = on_duplicate_candidates
        self.K = len(relevant)
        self.k = int(np.ceil(recall*self.K))
        #variable
        self.monitor = iter(monitor)
        self.delta_candidates = set()
        self.count = 0
        self.candidates = set()
        self.n_relevant = 0
        self.n_exhaustive_calls = 0
        self.n_delta_relevant = 0
        
    def sanitize_delta_candidates(self):
        """
        Helper function for SSWR
        """
        self.delta_candidates = set(self.delta_candidates)
        intersection = self.delta_candidates.intersection(self.candidates)
        if intersection and self.on_duplicate_candidates=='raise':
            msg = 'duplicate candidates found {}'
            raise RuntimeError(msg.format(intersection))
        self.delta_candidates = self.delta_candidates - intersection

    def __iter__(self):
        return self
    
    def __next__(self):
        self.delta_candidates = next(self.monitor)
        self.sanitize_delta_candidates()
        self.n_delta_relevant = len(self.delta_candidates.intersection(self.relevant))
        self.n_relevant += self.n_delta_relevant
        if self.k <= self.n_relevant:
            raise StopIteration()
        self.n_exhaustive_calls += len(self.delta_candidates)
        self.candidates.update(self.delta_candidates)
        
    def __call__(self, allow_halt=False):
        halt = self.k > self.n_relevant #is StopIteration()?
        if halt and not allow_halt:
            msg = ('with allow_halt=False, the delta_generator exited without '
                   'producing enough relevant documents. Needed {}, got {}')
            raise LookupError(msg.format(self.k, self.n_relevant))
        if halt:
            #random oracle search on the rest
            delta_N = self.N - self.n_exhaustive_calls
            delta_K = self.K - self.n_relevant
            delta_k = self.k - self.n_relevant
        else:
            #random oracle search on the last delta_candidates
            delta_N = len(self.delta_candidates)
            delta_K = self.n_delta_relevant
            delta_k = self.k - (self.n_relevant - self.n_delta_relevant)
        nume_work = MatchingOracleCost(delta_N, delta_K, delta_k)
        nume_work += self.n_exhaustive_calls + self.monitor.get_value()
        deno_work = MatchingOracleCost(self.N, self.K, self.k)
        work_ratio = nume_work / deno_work
        return (work_ratio, halt) if allow_halt else work_ratio
    
def SSWR(relevant, monitor, N, recall=1, allow_halt=False, on_duplicate_candidates='raise'):
    """
    The Sequential Search Work Ratio (SSWR) metric used for quick retrieval task
    this function use monitor has the delta generator and monitor.get_value() for the 
    the cost of generating all the previous delta candidates. On might consider using
    ChronoSSWR or CounterSSWR, which implement a precise monitor, instead.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    monitor : object
        Iterator of set of index and implementing get_value() -> float
    N : int
        The number of documents in the database
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    allow_halt : bool (optional)
        Allow the generator not to necessarily all needed indexes. This is similar, but
        not equal, to the case where the generator produce every other indexes at one
        and stop. (default False)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the duplicated candidate(s) will be removed. (default 'raise') 
    
    Returns
    -------
        work_ratio : float (or (float, bool) if allow_halt)
            By default, the SSWR. If allow_halt is True it returns a tuple with the
            SSWR and a boolean indicating if the generator halted abruptly.
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice.
    LookupError
        If allow_halt is False and delta_generator stops without
        generating enough relevant documents.
            
    Notes
    -----
    the aforementioned index is something hashable (i.e. hash(index) exists) that
    can be used to identify uniquely each document in the database. It might be the
    document itself, but that would be memory heavy. The most common case is using
    integer but in some cases it might be practical to have more information, maybe
    packed in a tuple for example.
    """
    sswr = _SSWR(relevant, monitor, N, recall=recall, on_duplicate_candidates=on_duplicate_candidates)
    for _ in sswr: pass
    return sswr(allow_halt=allow_halt)
        
def HaltSSWR(relevant, monitor, N, max_halt, recall=1, on_duplicate_candidates='raise'):
    """
    The Sequential Search Work Ratio (SSWR) metric used for quick retrieval task
    this function use monitor has the delta generator and monitor.get_value() for the 
    the cost of generating all the previous delta candidates. On might consider using
    ChronoSSWR or CounterSSWR, which implement a precise monitor, instead.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    monitor : object
        Iterator of set of index and implementing get_value() -> float
    N : int
        The number of documents in the database
    max_halt : int
        The maximum number of delta_candidates the generator is allow to give before halting.
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the duplicated candidate(s) will be removed. (default 'raise') 
    
    Returns
    -------
        sswrs : np.ndarray (shape=(max_halt+1), dtype=np.float)
            The SSWR w.r.t. all halting between 0 and max_halt. I.e. sswrs[i] is the
            sswr with halting i.
        halts : np.ndarray (shape=(max_halt+1), dtype=np.bool)
            halts[i] gives whether halting was used. I.e. if halts[i]==False implies
            all documents were found before the ith delta_candidates were produced
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice.
            
    Notes
    -----
    the aforementioned index is something hashable (i.e. hash(index) exists) that
    can be used to identify uniquely each document in the database. It might be the
    document itself, but that would be memory heavy. The most common case is using
    integer but in some cases it might be practical to have more information, maybe
    packed in a tuple for example.
    """
    sswrs = np.zeros((max_halt+1))
    halts = np.zeros((max_halt+1), dtype=np.bool)
    sswr = _SSWR(relevant, monitor, N, recall=recall, on_duplicate_candidates=on_duplicate_candidates)
    sswrs[0], halts[0] = sswr(allow_halt=True)
    n = 0 #if monitor is empty or find all relevants on first candidates
    for n, _ in enumerate(sswr,1):
        sswrs[n], halts[n] = sswr(allow_halt=True)
    sswrs[n+1:], halts[n+1:] = sswr(allow_halt=True)
    return sswrs, halts

def ChronoSSWR(relevant, delta_generator, N, eta=1, recall=1, allow_halt=False, on_duplicate_candidates='raise'):
    """
    The Chronometer Sequential Search Work Ratio (SSWR) metric used for quick retrieval task.
    It uses the generating time (in seconds) has a mesure of the work done by the delta_generator.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    delta_generator : generator of set of index
        This should generate set of candidates (index)
    N : int
        The number of documents in the database
    eta : positive float (optional)
        The proportion of importance between a generating time (seconds) and one oracle call.
        e.g. eta=2 implies that 1 generator second is equivalent to 2 oracle call. (default 1)
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    allow_halt : bool (optional)
        Allow the generator not to necessarily all needed indexes. This is similar, but
        not equal, to the case where the generator produce every other indexes at one
        and stop. (default False)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the diplicated candidate(s) will be removed. (default 'raise')
    
    Returns
    -------
        work_ratio : float (or (float, bool) if allow_halt)
            By default, the SSWR. If allow_halt is True it returns a tuple with the
            SSWR and a boolean indicating if the generator halted abruptly.
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice
    LookupError
        If delta_generator stops without generating enough relevant documents 
    """
    monitor = MonitorTime(delta_generator, eta=eta)
    return SSWR(relevant, monitor, N, recall=recall, allow_halt=allow_halt, on_duplicate_candidates=on_duplicate_candidates)

def CounterSSWR(relevant, delta_generator, N, eta=1, recall=1, allow_halt=False, on_duplicate_candidates='raise'):
    """
    The Counter Sequential Search Work Ratio (SSWR) metric used for quick retrieval task.
    It uses the number of call to the generator has a mesure of the work done by the delta_generator.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    delta_generator : generator of set of index
        This should generate set of candidates (index)
    N : int
        The number of documents in the database
    eta : positive float (optional)
        The proportion of importance between one generator call and one oracle call.
        e.g. eta=2 implies that 1 generator call is equivalent to 2 oracle call. (default 1)
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    allow_halt : bool (optional)
        Allow the generator not to necessarily all needed indexes. This is similar, but
        not equal, to the case where the generator produce every other indexes at one
        and stop. (default False)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the diplicated candidate(s) will be removed. (default 'raise')
    
    Returns
    -------
        work_ratio : float (or (float, bool) if allow_halt)
            By default, the SSWR. If allow_halt is True it returns a tuple with the
            SSWR and a boolean indicating if the generator halted abruptly.
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice
    LookupError
        If delta_generator stops without generating enough relevant documents 
    """
    monitor = MonitorNext(delta_generator, eta=eta)
    return SSWR(relevant, monitor, N, recall=recall, allow_halt=allow_halt, on_duplicate_candidates=on_duplicate_candidates)

def HaltingChronoSSWR(relevant, delta_generator, N, max_halt, eta=1, recall=1, on_duplicate_candidates='raise'):
    """
    The Chronometer Sequential Search Work Ratio (SSWR) metric used for quick retrieval task.
    It uses the number of call to the generator has a mesure of the work done by the delta_generator
    w.r.t to all possible halting from 0 to max_halt.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    delta_generator : generator of set of index
        This should generate set of candidates (index)
    N : int
        The number of documents in the database
    max_halt : int
        The maximum number of delta_candidates the generator is allow to give before halting.
    eta : positive float (optional)
        The proportion of importance between one generator call and one oracle call.
        e.g. eta=2 implies that 1 generator call is equivalent to 2 oracle call. (default 1)
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the diplicated candidate(s) will be removed. (default 'raise')
    
    Returns
    -------
        sswrs : np.ndarray (shape=(max_halt+1), dtype=np.float)
            The SSWR w.r.t. all halting between 0 and max_halt. I.e. sswrs[i] is the
            sswr with halting i.
        halts : np.ndarray (shape=(max_halt+1), dtype=np.bool)
            halts[i] gives whether halting was used. I.e. if halts[i]==False implies
            all documents were found before the ith delta_candidates were produced
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice
    """
    delta_generator = itertools.islice(delta_generator, 0, max_halt)
    monitor = MonitorTime(delta_generator, eta=eta)
    return HaltSSWR(relevant, monitor, N, max_halt, recall=recall, on_duplicate_candidates=on_duplicate_candidates)

def HaltingCounterSSWR(relevant, delta_generator, N, max_halt, eta=1, recall=1, on_duplicate_candidates='raise'):
    """
    The Counter Sequential Search Work Ratio (SSWR) metric used for quick retrieval task.
    It uses the number of call to the generator has a mesure of the work done by the delta_generator
    w.r.t to all possible halting from 0 to max_halt.
    
    Parameters
    ----------
    relevant : set of index
        Corresponding to the set of elemement that need to be retrieved
    delta_generator : generator of set of index
        This should generate set of candidates (index)
    N : int
        The number of documents in the database
    max_halt : int
        The maximum number of delta_candidates the generator is allow to give before halting.
    eta : positive float (optional)
        The proportion of importance between one generator call and one oracle call.
        e.g. eta=2 implies that 1 generator call is equivalent to 2 oracle call. (default 1)
    recall : float in [0,1] (optional)
        The minimal percentage of relevant document that should be generated
        (default 1)
    on_duplicate_candidates : str (optional)
        Should be 'raise' or 'ignore'. Set what should be done if the same index
        is generated twice. If 'raise', a RunetimeError will be raised. Otherwise,
        if 'ignore', the diplicated candidate(s) will be removed. (default 'raise')
    
    Returns
    -------
        sswrs : np.ndarray (shape=(max_halt+1), dtype=np.float)
            The SSWR w.r.t. all halting between 0 and max_halt. I.e. sswrs[i] is the
            sswr with halting i.
        halts : np.ndarray (shape=(max_halt+1), dtype=np.bool)
            halts[i] gives whether halting was used. I.e. if halts[i]==False implies
            all documents were found before the ith delta_candidates were produced
            
    Raises
    ------
    ValueError
        If on_duplicate_candidates not in {'raises', 'ignore'}.
    RuntimeError
        If on_duplicate_candidates=='raise' and an index is generated twice
    """
    delta_generator = itertools.islice(delta_generator, 0, max_halt)
    monitor = MonitorNext(delta_generator, eta=eta)
    return HaltSSWR(relevant, monitor, N, max_halt, recall=recall, on_duplicate_candidates=on_duplicate_candidates)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
with open("version.txt", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="radbm", # Replace with your own username
    version=version,
    author="Mathieu Duchesneau",
    author_email="duchesneau.mathieu@gmail.com",
    description="Rapid Access Database Benchmark",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/duchesneaumathieu/radbm",
    packages=setuptools.find_packages(),
    package_data={'radbm.utils.gdrive': ['fileid.json']},
    classifiers=[
        "Programming Language :: Python :: 3",
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy as np
from radbm.utils.time import Chronometer
from radbm.utils.stats import superdupergeometric_expectations

def oracle_expectation_by_recalls(recalls, keys, probs, values):
    """
    An utility function meant for a very specific usage, this
    compute the (maximum likelihood) expectation estimator for
    each recalls for values of a precise format.
    
    Parameters
    ----------
    recalls : iterable of float in [0,1]
        The recalls for which we want the expectations.
        
    keys : ndarray
        1D with dtype=int containing all (unique) possible number
        of relevant documents.
        
    probs : ndarray
        1D with len(probs)=len(keys) and sum(probs)==1. probs[i] is
        the estimated probability that the keys[i] is the number of
        relevant document. 
        
    values : dict (k int -> v list)
        With k the number of relevant document and v[i] is the estimated
        expectation of the oracle for the number of revelant document=k
        and a wanted recall of i/k.
    
    Returns
    -------
        expectations : ndarray
            1D with dtype=float. expectations[i] is the expectation for an
            oracle stopping at a recall of at least recalls[i].
    """
    expectations = list()
    for recall in recalls:
        idxs = np.ceil(keys*recall).astype(np.int)
        e = sum(p*values[k][i] for i,k,p in zip(idxs,keys,probs))
        expectations.append(e)
    return np.array(expectations)

def oracle_random_calls_expectation_by_recall(recalls, keys, probs, N):
    """
    This function gives the expected number of calls (by query) to an oracle we
    would need to make to get a desired recall by trying documents at random.
    
    This is a baseline to any algorithm producing an order over the documents for
    the oracle to try. An algorithm producing an order on which the oracle would
    need (in expectation) more calls than what is returned by this function 
    [to achieve the same recall] as no value [for this particular recall].
    
    Parameters
    ----------
    recalls : numpy.ndarray
        The recalls for which we want the expected numbers
        
    keys : ndarray
        1D with dtype=int containing all (unique) possible number
        of relevant documents.
        
    probs : ndarray
        1D with len(probs)=len(keys) and sum(probs)==1. probs[i] is
        the estimated probability that the keys[i] is the number of
        relevant document. 
        
    N : int
        The total number of documents
    Returns
    -------
    expectations : numpy.ndarray
        The expected number of oracle calls for each recall given as input
        
    Notes
    -----
    How it works: For each query we have K relevant document out of N, the probability
    that the oracle find k relevant document in n samples is given by the superdupergeometric
    distribution SDGD. 
    This algorithm estimate (with maximum likelihood) the probability that their
    is K relevant documents for a query. Than for a specific recall r and for
    each k (with estimated pmf(K=k) > 0) we can find the mininum number of relevant
    documents to be found the get at least r of recall. 
    Finaly we estimate this expectation for a specific recall r with
    sum_{k s.t. pmf(K=k)>0} pmf(K=k) E[SDGD(N,k)[:, k_r]]
    """    
    expect_by_k = {k:superdupergeometric_expectations(N, k) for k in keys}
    expectations = oracle_expectation_by_recalls(recalls, keys, probs, expect_by_k)
    return expectations

def candidates_time_oracle_calls_count(candidates_iterator, relevant_set, recall_break=1):
    breaking_number = int(np.ceil(recall_break*len(relevant_set)))
    times = np.zeros((breaking_number+1,))
    calls = np.zeros((breaking_number+1,))
    calls_counter = 0
    founds_set = set()
    chronometer = Chronometer().start()
    for c in candidates_iterator:
        chronometer.stop()
        calls_counter += 1
        if c in relevant_set and not c in founds_set:
            founds_set.add(c)
            times[len(founds_set)] = chronometer.time()
            calls[len(founds_set)] = calls_counter
            if len(founds_set) == breaking_number:
                break #if candidates order is good measuring will be fast
        chronometer.start()
    if len(founds_set) != breaking_number:
        raise RuntimeError('Model did not found every document needed')
    return times, calls

def iter_candidates(model, query):
    for candidates in model.candidates(query):
        for candidate in candidates:
            yield candidate

def run_oracle_experiments(model, query_iterator, N, recalls=[1]):
    """
    This function computes the ToCR for each value in recalls.
    
    Parameters
    ----------
    model : IRModel
        model.candidates must be implemented and the model must
        already be indexing a set of documents.
        
    query_iterator : iterator
        (query, relevant_set) <- next(query_iterator) with query
        compatible with model.candidates and relevant_set containing
        the id of the relevant documents in the index of the model.
        
    N : int
        The number of documents in the index.
        
    recalls : iterable of float in [0, 1]
        The recalls for which we want to compute the ToCR.
        
    Returns
    -------
    expected_times : ndarray
        1D, shape=(len(recalls),), dtype=float. expected_times[i] correspond to
        the expected time it takes to iterate over enough candidates
        to get recall=recalls[i].
        
    expected_calls : ndarray
        1D, shape=(len(recalls),), dtype=float. expected_calls[i] correspond to
        the expected number of oracle calls it takes to iterate over enough
        candidates to get recall=recalls[i].
        
    expected_calls_baseline : ndarray
        1D, shape=(len(recalls),), dtype=float. expected_calls_baseline[i]
        correspond to the expected number of oracle calls it would take if
        we iterate randomly over all documents indexed to get recall=recalls[i].
        
     """
    times_dict = dict()
    calls_dict = dict()
    probs = dict()
    keys = list()
    recall_break = max(recalls)
    for query, relevant_set in query_iterator:
        candidates_iterator = iter_candidates(model, query)
        times, calls = candidates_time_oracle_calls_count(
            candidates_iterator, relevant_set, recall_break)
        k = len(relevant_set)
        if k in times_dict:
            times_dict[k].append(times)
            calls_dict[k].append(calls)
            probs[k] += 1
        else:
            keys.append(k)
            times_dict[k] = [times]
            calls_dict[k] = [calls]
            probs[k] = 1
            
    keys = np.array(keys)
    probs = np.array([probs[k] for k in keys])/sum(probs.values())
    times_dict = {k:np.array(v).mean(axis=0) for k,v in times_dict.items()}
    calls_dict = {k:np.array(v).mean(axis=0) for k,v in calls_dict.items()}
    
    expected_times = oracle_expectation_by_recalls(recalls, keys, probs, times_dict)
    expected_calls = oracle_expectation_by_recalls(recalls, keys, probs, calls_dict)
    expected_calls_baseline = oracle_random_calls_expectation_by_recall(recalls, keys, probs, N)
    
    return expected_times, expected_calls, expected_calls_baseline
        
def time_over_oracle_calls_reduction_metric(model, query_iterator, N, recalls=[1]):
    """
    This function computes the ToCR for each value in recalls.
    
    Parameters
    ----------
    model : IRModel
        model.candidates must be implemented and the model must
        already be indexing a set of documents.
        
    query_iterator : iterator
        (query, relevant_set) <- next(query_iterator) with query
        compatible with model.candidates and relevant_set containing
        the id of the relevant documents in the index of the model.
        
    N : int
        The number of documents in the index.
        
    recalls : iterable of float in [0, 1]
        The recalls for which we want to compute the ToCR.
        
    Returns
    -------
    ToCRs : ndarray
        1D, shape=(len(recalls),), dtype=float. ToCRs[i] correspond to
        the ToCR with recall=recalls[i].
     """
    exp = run_oracle_experiments(model, query_iterator, N, recalls=recalls)
    expect_times, expect_calls, expect_calls_baseline = exp
    expect_calls_reduction = expect_calls_baseline - expect_calls
    
    strictly_positive = 0 < expect_calls_reduction
    safe_div = np.where(strictly_positive, expect_calls_reduction, 1)
    ToCRs = np.where(strictly_positive, expect_times/safe_div, np.inf)
    return ToCRs


import unittest, torch
import numpy as np
from radbm.search.elba import HashNet
from radbm.search.mbsds import HashingMultiBernoulliSDS

class TestHashNet(unittest.TestCase):
    def test_step(self):
        torch.manual_seed(0)
        model = HashNet(
            torch.nn.Linear(32,64),
            torch.nn.Linear(32,64),
            HashingMultiBernoulliSDS(1,1),
            1/32,
        )
        x = torch.eye(32, dtype=torch.float)
        r = torch.eye(32, dtype=torch.bool)
        l0 = model.step(x, x, r)
        for i in range(500):
            model.step(x, x, r)
        l1 = model.step(x, x, r)
        self.assertTrue(l1 < l0)


import unittest
import numpy as np
from radbm.search.gridsearch import GridSearch

class TestGridSearch(unittest.TestCase):
    def setUp(self):
        self.bounds = np.array([[-1,-2,-3], [3,2,1]])
        self.divisions = np.array([4, 2, 8])
        self.gs = GridSearch(self.bounds, self.divisions)
    
    def test_point_arithmetic(self):
        points = np.array([
            [0.6, 1.5, -2.1],
            [0, 1, 0],
        ])
        scaled_points = self.gs.compute_scaled_points(points)
        cells = self.gs.compute_cells(points)
        pos = self.gs.compute_cell_relative_positions(points)
        self.assertTrue((pos==(scaled_points-cells)).all())
        
        self.assertTrue(self.gs.is_cell_inbound((0,0,0)))
        self.assertFalse(self.gs.is_cell_inbound((0,2,0)))
        
        self.assertTrue(self.gs.is_points_inbound(np.array([0,0,0])))
        self.assertTrue(self.gs.is_points_inbound(np.array([-1,-2,-3])))
        self.assertFalse(self.gs.is_points_inbound(np.array([-4,6,2])))
        self.assertFalse(self.gs.is_points_inbound(np.array([3,0,0])))
        
        #graphsearch_cell_generator specific
        min_bounds, max_bounds = self.bounds
        width = (max_bounds-min_bounds)/self.divisions
        self.assertEqual(list(width**2), list(self.gs.width2))
        delta_x = np.stack([pos[0], 1-pos[0]])
        rel_cells = np.array([
            [0,0,0],
            [1,0,0],
            [3,0,0],
            [0,0,2],
            [0,0,-2],
            [3,0,-2],
        ])
        expected_dists2 = [
            0,
            (delta_x[1,0]*width[0])**2,
            ((delta_x[1,0] + 2)*width[0])**2,
            ((delta_x[1,2] + 1)*width[2])**2,
            ((delta_x[0,2] + 1)*width[2])**2,
            ((delta_x[1,0] + 2)*width[0])**2 + ((delta_x[0,2] + 1)*width[2])**2,
        ]
        dists2 = self.gs.point_relative_cells_dists(delta_x, rel_cells)**2
        self.assertEqual(expected_dists2, list(dists2))
        
        #bruteforce_cell_generator specific
        #implicitly tested
        #self.gs.compute_cells_dist(self.points[0])
        
    def test_cell_generators(self):
        point = np.array([0.12, -1.1, .74])
        g1 = list(self.gs.graphsearch_cell_generator(point))
        g2 = list(self.gs.bruteforce_cell_generator(point))
        self.assertEqual(g1, g2)
        
    def test_retrieval(self):
        #out of bound point
        out_point = np.array([-4,0,0])
        with self.assertRaises(ValueError):
            self.gs.insert(out_point, 1)
            
        with self.assertRaises(ValueError):
            next(self.gs.itersearch(out_point))
            
        rng = np.random.RandomState(0xcafe)
        min_bounds, max_bounds = self.bounds
        delta = max_bounds - min_bounds
        
        #test double insert
        points1 = rng.uniform(0,1,(1000,3))*delta + min_bounds
        self.gs.batch_insert(points1, range(1000))
        points2 = rng.uniform(0,1,(1000,3))*delta + min_bounds
        self.gs.batch_insert(points2, range(1000,2000))
        points = np.concatenate([points1, points2])
        
        point = np.array([0.12, -1.1, .74])
        dists2 = ((point-points)**2).sum(axis=1)
        argsort = list(np.argsort(dists2))
        
        #ndim!=1 error
        with self.assertRaises(ValueError):
            next(self.gs.itersearch(point[None]))
        
        #wrong cell_generator error
        with self.assertRaises(ValueError):
            next(self.gs.itersearch(point, cell_generator='something'))
        
        #test exactness
        graphsearch = list(self.gs.itersearch(
            point, cell_generator='graphsearch', 
            max_dist=np.inf, max_items=np.inf
        ))
        self.assertEqual(graphsearch, argsort)
        
        bruteforce = list(self.gs.itersearch(
            point, cell_generator='bruteforce', 
            max_dist=np.inf, max_items=np.inf
        ))
        self.assertEqual(bruteforce, argsort)
        
        #test max_items==0
        self.assertEqual(list(self.gs.itersearch(point, max_items=0)), list())
        
        #test max_dist==0
        self.assertEqual(list(self.gs.itersearch(point, max_dist=0)), list())
        
        #test max_items>0
        self.assertEqual(len(list(self.gs.itersearch(point, max_items=42))), 42)
        
        #test max_dist>0
        items = list(self.gs.itersearch(point, max_dist=1))
        self.assertTrue(np.sqrt(dists2[items[-1]]) < 1)
        self.assertTrue(np.sqrt(dists2[argsort[np.where(argsort==items[-1])[0][0]+1]] >= 1))
        
        #test max_items>0 and all cell generated
        self.assertEqual(len(list(self.gs.itersearch(point, max_items=1999))), 1999)


import torch
import numpy as np

def hamming_distance(x, y, *args, **kwargs):
    """
    Compute the Hamming distance.
    
    Parameters
    ----------
    x : torch.Tensor (dtype=torch.bool)
    y : torch.Tensor (dtype=torch.bool)
    *args
        Passed to sum
    *kwargs
        Passed to sum
    
    Returns
    -------
    z : torch.Tensor (dtype=torch.int64)
        The Hamming distance between x and y
    """
    
    z = (x^y).sum(*args, **kwargs)
    return z

def conditional_hamming_counts(documents, queries, relevances, batch_size=100):
    """
    Compute the count of relevant and not relevant match w.r.t all Hamming distance.
    
    Parameters
    ----------
    documents : torch.Tensor (2D, dtype=torch.bool)
        The binary reprentation of a batch of documents (database).
        documents[i] is the ith document. documents.shape[1] must be 
        equal to queries.shape[1]. (should also be on the same device as queries)
    queries : torch.Tensor (2D, dtype=torch.bool)
        The binary reprentation of a batch of queries. queries[i]
        is the ith query. queries.shape[1] must be equal to
        documents.shape[1]. (should also be on the same device as documents)
    relevances : list of set of int
        len(relevances) must be len(queries). For each corresponding query,
        it give the set of relevant documents (given by its index). Explicitly,
        j in relevances[i] iff query[i] matches with documents[j].
    batch_size : int (optional)
        The number of query for which we compute the Hamming distance at a time.
        if it is to big the results might not fit in RAM (or on the GPU). (default 100)
        
    Returns
    -------
    relevant_counts : torch.Tensor (1D, dtype=torch.float)
        len(relevant_counts) = queries.shape[1] + 1 (also equal to documents.shape[1] + 1) and
        relevant_dcounts[i] is the number of relevant documents at Hamming distance i.
    irrelevant_counts : torch.Tensor (1D, dtype=torch.float)
        len(irrelevant_countst) = queries.shape[1] + 1 (also equal to documents.shape[1] + 1) and
        irrelevant_counts[i] is the number of irrelevant documents be at Hamming distance i.
        
    Notes
    -----
    This assume that each sets in relevances is small compared to len(documents) and should
    be used on a GPU otherwise it is quite slow.
    """
    m, n = queries.shape
    if n != documents.shape[1]:
        msg = 'queries and documents binary codes are not of the same lenght, got {} and {} respectively.'
        raise ValueError(msg.format(n, documents.shape[1]))
    if m != len(relevances):
        msg = 'len(queries) != len(relevances), got {} != {}.'
        raise ValueError(msg.format(m, len(relevances)))
    
    device = queries.device
    nbatch = int(np.ceil(m/batch_size))
    total_counts = torch.zeros(n+1, device=device, dtype=torch.int64)
    relevant_counts = torch.zeros(n+1, device=device, dtype=torch.int64)
    for i in range(nbatch):
        a, b = i*batch_size, (i+1)*batch_size
        dists = hamming_distance(queries[a:b,None], documents[None], dim=2)
        
        #update total_count
        udists, ucounts = torch.unique(dists, return_counts=True)
        total_counts[udists] += ucounts
        
        #update relevant_count
        rows, cols = zip(*((i,j) for i, rel in enumerate(relevances[a:b]) for j in rel))
        udists, ucounts = torch.unique(dists[(rows,cols)], return_counts=True)
        relevant_counts[udists] += ucounts
    
    irrelevant_counts = total_counts - relevant_counts
    return relevant_counts, irrelevant_counts

def conditional_counts_to_pr_curve(relevant_counts, irrelevant_counts, return_valid_dists=False):
    total_counts = relevant_counts + irrelevant_counts
    tcs = total_counts.cumsum(dim=0)
    rcs = relevant_counts.cumsum(dim=0)
    recalls = rcs.float() / sum(relevant_counts)
    dists = torch.where(tcs != 0)[0]
    precisions = (torch.zeros_like(recalls)-1).log() #nans_like
    precisions[dists] = rcs[dists].float()/tcs[dists]
    if return_valid_dists:
        return dists, precisions[dists], recalls[dists]
    return precisions, recalls

def hamming_pr_curve(documents, queries, relevances, batch_size=100, return_valid_dists=False):
    """
    Compute the precision-recall curve w.r.t the Hamming distance. I.e. it
    computes the precision and recall for each Hamming distance decision
    thresholds.  
    
    Parameters
    ----------
    documents : torch.Tensor (2D, dtype=torch.bool)
        The binary reprentation of a batch of documents (database).
        documents[i] is the ith document. documents.shape[1] must be 
        equal to queries.shape[1]. (should be on the same device as queries)
    queries : torch.Tensor (2D, dtype=torch.bool)
        The binary reprentation of a batch of queries. queries[i]
        is the ith query. queries.shape[1] must be equal to
        documents.shape[1]. (should be on the same device as documents)
    relevances : list of set of int
        len(relevances) must be len(queries). For each corresponding query,
        it give the set of relevant documents (given by its index). Explicitly,
        j in relevances[i] iff query[i] matches with documents[j].
    batch_size : int (optional)
        The number of query for which we compute the Hamming distance at a time.
        if it is to big the results might not fit in RAM (or on the GPU). (default 100)
    return_valid_dists : bool (optional)
        Some dists might have an undefined precision. In those case the returned value 
        will be nan by default. If return_valid_dists is True those value won't be there
        and dists will be returned with precision and recall. 
        See the returns section for more info. (default False)
        
    Returns
    -------
    dists : torch.Tensor (1D, dtype=torch.int64) if return_valid_dists is True
        Only present if return_valid_dists is True. It correspond to the valid distances
        where the precision is define.
    precisions : torch.Tensor (1D, dtype=torch.float)
        len(precisions) = queries.shape[1] + 1 (also equal to documents.shape[1] + 1) and
        precision[i] is the precison w.r.t. a Hamming distance of i if return_valid_dists
        is False otherwise, len(precisions) = len(dists) and precisions[i] is the precision
        w.r.t a Hamming distance of dists[i].
    recalls : torch.Tensor (1D, dtype=torch.float)
        len(recalls) = queries.shape[1] + 1 (also equal to documents.shape[1] + 1) and
        recalls[i] is the recall w.r.t. to a Hamming distance of i if return_valid_dists
        is False otherwise, len(recalls) = len(dists) and recalls[i] is the recall
        w.r.t a Hamming distance of dists[i].
        
    Notes
    -----
    This assume that each sets in relevances is small compared to len(documents) and should
    be used on a GPU otherwise it is quite slow.
    """
    relevant_counts, irrelevant_counts = conditional_hamming_counts(
        documents, queries, relevances, batch_size=batch_size)
    return conditional_counts_to_pr_curve(
        relevant_counts, irrelevant_counts, return_valid_dists=return_valid_dists)


import numpy as np
from IPython.core.display import display, clear_output, HTML

html_caption = '<caption style="text-align:left">{}</caption>'
html_table = '<table style="float:left;width:50%">{}</table>'
html_row = '<tr>{}</tr>'
html_item = '<th style="text-align:center">{}</th>'
html_bold_item = '<th style="text-align:center"><b>{}</b></th>'

def notebook_html_table(table, title=None, cols_name=None, rows_name=None):
    rows = list()
    for n, row in enumerate(table):
        row = ['{:.4f}'.format(e) for e in row]
        row = list(map(html_item.format, row))
        if rows_name: row = [html_bold_item.format(rows_name[n])] + row
        rows.append(html_row.format(''.join(row)))
    if cols_name:
        if rows_name: cols_name = [''] + cols_name
        header = list(map(html_bold_item.format, cols_name))
        header = html_row.format(''.join(header))
        rows = [header] + rows
    if title: rows = [html_caption.format(title)] + rows
    table = html_table.format(''.join(rows))
    clear_output(wait=True)
    display(HTML(table))
    return table


import torch #only used in cast_numpy
import numpy as np
from radbm.search.base import BaseSDS
from radbm.utils.stats.generators import greatest_k_multi_bernoulli_outcomes_generator

def cast_numpy(data):
    if isinstance(data, torch.Tensor):
        return data.detach().cpu().numpy()
    return data

class HashingMultiBernoulliSDS(BaseSDS):
    """
    This algorithm build N hash tables from documents in the form
    of Multi-Bernoulli distributions. From a documents (i.e. a
    distribution) we find the i-th most probable outcome (which is
    a binary vector) and use it as a key for the corresponding
    i-th hash table.
    
    For retrieval, a query (i.e. a Multi-Bernoulli distribution) 
    we use the M most probable outcomes (which are binary vectors)
    to do M lookups in the hash tables. In this context, a lookups
    means looking in each tables. In total NxM hash table calls
    will be made.
    
    Parameters
    ----------
    ntables : int
        The number of tables to build
    nlookups : int, optional (default=1)
        The number of lookups to do when searching
    
    Notes
    -----
        The Multi-Bernoulli distribution are always parametrized with the
        vector of log probabilities for the bits to be one.
    """
    def __init__(self, ntables, nlookups=1):
        self.ntables = ntables
        self.nlookups = nlookups
        self.reset()
        
    def reset(self):
        """
        Empty the hash tables
        
        Returns
        -------
        self
        """
        self.tables = [dict() for _ in range(self.ntables)]
        return self
        
    def get_buckets_avg_size(self):
        """
        Returns
        -------
        buckets_avg_size : list of float
            The average number of documents per buckets for each
            hash tables
        """
        return [np.mean(list(map(len, t.values()))) for t in self.tables]
    
    def get_buckets_max_size(self):
        """
        Returns
        -------
        buckets_max_size : list of float
            The maximum number of documents per buckets for each
            hash tables
        """
        return [np.max(list(map(len, t.values()))) if t else 'nan' for t in self.tables]
        
    def __repr__(self):
        r = 'Search: {}\nTables size: {}\nBuckets avg size: {}\nBuckets max size: {}'
        return r.format(
            self.__class__.__name__,
            ', '.join(map(str, map(len, self.tables))),
            ', '.join(['{:.2f}'.format(s) for s in self.get_buckets_avg_size()]),
            ', '.join(map(str, self.get_buckets_max_size())),
        )
    
    def _get_generator(self, log_probs, k):
        """
        Parameters
        ----------
        log_probs : numpy.ndarray (ndim == 1 or 2)
            If ndim==1: The Multi-Bernoulli distribution parametrized in log probabilities
            If ndim==2: len(log_probs) should be 2. The first element should be log
            probabilities that bits are zero and the second element should be the log
            probabilities that bits are one. This is for numerical stability
            (i.e. when probability are too close to 1)
        k : int
            The number of outcomes to generates
        """
        log_probs = cast_numpy(log_probs)
        if log_probs.ndim==1:
            return greatest_k_multi_bernoulli_outcomes_generator(log_probs, k=k)
        elif log_probs.ndim==2:
            log_probs0, log_probs1 = log_probs
            return greatest_k_multi_bernoulli_outcomes_generator(log_probs0, log_probs1, k=k)
        else:
            msg = 'log_probs.ndim should be 1 or 2, got {}'
            raise ValueError(msg.format(log_probs.ndim))
        
    def insert(self, log_probs, i):
        """
        Insert a unique document's index in the each tables. The document
        most be a Multi-Bernoulli distribution parametrized in log
        probabilities.
        
        Parameters
        ----------
        log_probs : numpy.ndarray (ndim == 1 or 2)
            If ndim==1: The Multi-Bernoulli distribution parametrized in log probabilities
            If ndim==2: len(log_probs) should be 2. The first element should be log
            probabilities that bits are zero and the second element should be the log
            probabilities that bits are one. This is for numerical stability
            (i.e. when probability are too close to 1)
        i : hashable (e.g. int or tuple)
            The index of the document
        """
        gen = self._get_generator(log_probs, self.ntables)
        for n, bits in enumerate(gen):
            table = self.tables[n]
            if bits in table: table[bits].add(i)
            else: table[bits] = {i}
                
    def search(self, log_probs, nlookups=None):
        """
        Search in the tables with a query in the form of a Multi-Bernoulli
        distribution parametrized in log probabilities. This will search in
        each tables with each of the top (nlookups) outcomes.
        
        Parameters
        ----------
        log_probs : numpy.ndarray (ndim == 1 or 2)
            If ndim==1: The Multi-Bernoulli distribution parametrized in log probabilities
            If ndim==2: len(log_probs) should be 2. The first element should be log
            probabilities that bits are zero and the second element should be the log
            probabilities that bits are one. This is for numerical stability
            (i.e. when probability are too close to 1)
        nlookups : int, optional
            The number of top outcome to uses for searching. If not specified
            the default self.nlookups is used.
            
        Returns
        -------
        indexes : set
            The indexes of each documents found in the search
        """
        indexes = set()
        nlookups = self.nlookups if nlookups is None else nlookups
        #loop for nlookups*self.ntables minus the number of empty lookups
        for new_indexes in self.itersearch(log_probs, nlookups):
            indexes.update(new_indexes)
        return indexes
    
    def itersearch(self, log_probs, nlookups=None, yield_empty=False):
        """
        Generator that search in the tables with a query in the form of a 
        Multi-Bernoulli distribution parametrized in log probabilities.
        This will search in each tables with each of the top (nlookups)
        outcomes. Everytime a set of indexes is found, this generator will
        yield it. If yield_empty is True, empty set will also be yield.

        Parameters
        ----------
        log_probs : numpy.ndarray (ndim == 1 or 2)
            If ndim==1: The Multi-Bernoulli distribution parametrized in log probabilities
            If ndim==2: len(log_probs) should be 2. The first element should be log
            probabilities that bits are zero and the second element should be the log
            probabilities that bits are one. This is for numerical stability
            (i.e. when probability are too close to 1)
        nlookups : int, optional
            The upper limit to generate the next most probable outcomes. Not to
            be confused with the number of item generated. By default, generates
            every outcomes.

        Yields
        ------
        indexes : set
            The newly found indexes
        """
        yielded = set()
        for bits in self._get_generator(log_probs, nlookups):
            for table in self.tables:
                if bits in table:
                    indexes = table[bits]
                    new_indexes = indexes - yielded
                    yielded.update(new_indexes)
                    if yield_empty or new_indexes:
                        yield new_indexes
                elif yield_empty:
                    yield set()
    
    def get_state(self):
        """
        Returns
        -------
        tables : list of dict
            The current hash tables
        """
        return self.tables
    
    def set_state(self, state):
        """
        Parameters
        -------
        state : list of dict
            The hash tables to use
        """
        self.tables = state
        return self


import unittest
import numpy as np
from radbm.utils import unique_list, Ramp

class Test_unique_list(unittest.TestCase):
    def test_unique_list(self):
        it = [3,4,4,2,1,1,0,3,2]
        expected_list = [3,4,2,1,0]
        result = unique_list(it)
        self.assertEqual(result, expected_list)
        
class TestRamp(unittest.TestCase):
    def test_ramp(self):
        #from x=3..5 we ramp from y=-1..7
        ramp = Ramp(3, 5, -1, 7)
        xs = np.array([2, 3, 4, 5, 6])
        ys = np.array([-1, -1, 3, 7, 7])
        self.assertTrue(np.allclose(ramp(xs), ys))


import torch, unittest
import numpy as np
from radbm.loaders.base import Loader, IRLoader

class DummyLoader(Loader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        #registering each dataset to switch
        #this is only possible if the datasets are small
        self.register_switch('train_dataset', 1*np.ones((2,3)))
        self.register_switch('valid_dataset', 2*np.ones((2,3)))
        self.register_switch('test_dataset', 3*np.ones((2,3)))
        
        #updating <which> groups.
        #that way, e.g. when self.valid() will be called
        #self.dataset will contain self.valid_dataset
        self.train_group['dataset'] = self.train_dataset
        self.valid_group['dataset'] = self.valid_dataset
        self.test_group['dataset'] = self.test_dataset
        self.end_init()
    
class TestLoader(unittest.TestCase):
    def setUp(self):
        self.dummy = DummyLoader('train', 'numpy', 'cpu')
        
    def try_switch_cuda(self):
        if torch.cuda.is_available():
            self.dummy.cuda()
            self.assertCUDA()
        else:
            with self.assertRaises(ValueError):
                self.dummy.cuda()
        
    def test_init(self):
        #test value_check
        with self.assertRaises(ValueError):
            self.dummy = DummyLoader('training', 'numpy', 'cpu')
        
        self.dummy = DummyLoader('train', 'numpy', 'cpu')
        self.assertDatasetValue(1)
        self.assertNumpy()
        
        self.dummy = DummyLoader('valid', 'numpy', 'cpu')
        self.assertDatasetValue(2)
        self.assertNumpy()
        
        #test numpy + cuda init error
        with self.assertRaises(ValueError):
            self.dummy = DummyLoader('valid', 'numpy', 'cuda')
        
        self.dummy = DummyLoader('valid', 'torch', 'cpu')
        self.assertDatasetValue(2)
        self.assertTorch()
        self.assertCPU()
        
        if torch.cuda.is_available():
            self.dummy = DummyLoader('test', 'torch', 'cuda')
            self.assertDatasetValue(3)
            self.assertTorch()
            self.assertCUDA()
        else:
            with self.assertRaises(ValueError):
                self.dummy = DummyLoader('test', 'torch', 'cuda')

        self.dummy.cpu().numpy()
        
        #test register_switch 'already exists' error
        with self.assertRaises(ValueError):
            self.dummy.register_switch('dataset', np.array(0))
        
    def assertNumpy(self):
        self.assertTrue(isinstance(self.dummy.dataset.data, np.ndarray))
        
    def assertTorch(self):
        self.assertTrue(isinstance(self.dummy.dataset.data, torch.Tensor))
        
    def assertCPU(self):
        self.assertTrue(self.dummy.dataset.data.device.type=='cpu')
        
    def assertCUDA(self):
        self.assertTrue(self.dummy.dataset.data.device.type=='cuda')
        
    def test_backend_and_device(self):
        #test init backend
        self.assertNumpy()
        
        #test cuda switch error
        with self.assertRaises(ValueError):
            self.dummy.cuda()
        
        #test switch to torch
        self.dummy.torch()
        self.assertTorch()
        self.assertCPU()
        
        #test switch to CUDA
        if torch.cuda.is_available():
            self.dummy.cuda()
            self.assertCUDA()

            #test numpy switch error
            with self.assertRaises(ValueError):
                self.dummy.numpy()
        else:
            with self.assertRaises(ValueError):
                self.dummy.cuda()
        
        #test switch to CPU
        self.dummy.cpu()
        self.assertCPU()
        
        #test switch to (back) numpy
        self.dummy.numpy()
        self.assertNumpy()
        
    def assertDatasetValue(self, value):
        eq = self.dummy.dataset.data==value
        self.assertTrue(eq.all())
    
    def _test_which(self):
        self.dummy.train()
        self.assertDatasetValue(1)
        
        self.dummy.valid()
        self.assertDatasetValue(2)
        
        self.dummy.test()
        self.assertDatasetValue(3)
    
    def test_which(self):
        self._test_which()
        
        self.dummy.torch()
        self._test_which()
        
        self.try_switch_cuda()
        self._test_which()
        
        self.dummy.cpu().numpy()
        self._test_which()
        
    def test_dynamic_cast(self):
        data = torch.tensor(0, device='cpu')
        
        #torch to numpy
        data = self.dummy.dynamic_cast(data)
        self.assertTrue(isinstance(data, np.ndarray))
        
        self.dummy.torch()
        data = self.dummy.dynamic_cast(data)
        self.assertTrue(isinstance(data, torch.Tensor))
        self.assertTrue(data.device.type=='cpu')
        
        if torch.cuda.is_available():
            self.dummy.cuda()
            data = self.dummy.dynamic_cast(data)
            self.assertTrue(isinstance(data, torch.Tensor))
            self.assertTrue(data.device.type=='cuda')
        
        self.dummy.cpu()
        data = self.dummy.dynamic_cast(data)
        self.assertTrue(isinstance(data, torch.Tensor))
        self.assertTrue(data.device.type=='cpu')
        
        self.dummy.numpy()
        
    def test_utilities(self):
        repr(self.dummy) #make sure it does not crash
        
        state = self.dummy.get_state()
        self.dummy.set_state(state)
        
        rng = self.dummy.get_rng()
        self.dummy.set_rng(rng) #numpy.random.RandomState
        self.dummy.set_rng(self.dummy.rng) #TorchNumpyRNG
        
    def test_without_cuda(self):
        is_available = torch.cuda.is_available
        torch.cuda.is_available = lambda : False
        try:
            #retest everything without cuda
            self.test_init()
            self.test_backend_and_device()
            self.test_which()
            self.test_dynamic_cast()
            self.test_utilities()
        finally:
            torch.cuda.is_available = is_available
            
class TestIRLoader(unittest.TestCase):
    def test_utilities(self):
        modes = IRLoader.get_available_modes(None)
        mode = next(iter(modes))
        dummy = IRLoader(mode, 'train', 'numpy', 'cpu')
        repr(dummy)